import {
  ConnectedPosition,
  FlexibleConnectedPositionStrategy, FlexibleConnectedPositionStrategyOrigin,
  Overlay,
  OverlayPositionBuilder,
  OverlayRef
} from '@angular/cdk/overlay';
import { TemplatePortal } from '@angular/cdk/portal';
import {
  ChangeDetectorRef,
  Directive,
  ElementRef,
  Inject,
  Input,
  NgZone,
  OnDestroy,
  Optional,
  TemplateRef,
  ViewContainerRef
} from '@angular/core';
import { Subscription } from 'rxjs';
import { NavigationStart, Router } from '@angular/router';
import { filter } from 'rxjs/operators';
import { POPOVER_OPTIONS } from './monkeyecx-popover-options.directive';
import type { PopoverDir, MonkeyEcxPopoverOptions } from './monkeyecx-popover-options.directive';

@Directive({
  selector: '[monkeyecxPopover]'
})
export class MonkeyEcxPopoverDirective implements OnDestroy {
  private _popover!: boolean;

  get popover() {
    return this._popover;
  }

  @Input('monkeyecxPopover')
  set popover(show: boolean) {
    this._popover = show;
    if (show) {
      this.createPopover();
    } else if (this.overlayRef) {
      this.overlayRef.dispose();
    }
  }

  @Input('monkeyecxPopoverClosed') closed!: (e?: MouseEvent) => void;

  private _target!: HTMLElement;

  get target() {
    return this._target;
  }

  @Input('monkeyecxPopoverTarget')
  set target(target: HTMLElement) {
    this._target = target;
  }

  @Input('monkeyecxPopoverMinwidth') minwidth = false;

  @Input('monkeyecxPopoverBackdrop') backdrop = true;

  @Input('monkeyecxPopoverWatch') watch = false;

  @Input('monkeyecxPopoverDir') dir: PopoverDir = 'ltr';

  @Input('monkeyecxPopoverContextmenu') contextmenu = false;

  @Input('monkeyecxPopoverHeight') height!: number | string;

  private overlayRef!: OverlayRef;

  private positionStrategy!: FlexibleConnectedPositionStrategy;

  private templatePortal!: TemplatePortal;

  private subs = new Subscription();

  private eventsSubs = new Subscription();

  private resizeObserver: any;

  constructor(
     private tpl: TemplateRef<any>,
     private vcr: ViewContainerRef,
     private el: ElementRef,
     private zone: NgZone,
     private _viewContainerRef: ViewContainerRef,
     private _cd: ChangeDetectorRef,
     private overlay: Overlay,
    private overlayPositionBuilder: OverlayPositionBuilder,
     private router: Router,
    @Inject(POPOVER_OPTIONS) @Optional() private options: MonkeyEcxPopoverOptions
  ) {
    this.eventsSubs = this.router.events
      .pipe(
        filter((event: any) => {
          return event instanceof NavigationStart;
        })
      )
      .subscribe(() => {
        this.onHandleClose();
      });
  }

  private onHandleClose(e?: MouseEvent) {
    this.popover = false;
    this.subs.unsubscribe();
    this.resizeObserver?.disconnect();
    this.zone.run(() => {
      if (this.closed) {
        this.closed(e);
      }
    });
  }

  getDirPositions(dir?: PopoverDir): ConnectedPosition[] {
    const revert = {
      start: 'end', end: 'start', bottom: 'top', top: 'bottom', center: 'center'
    };
    const centerY = (pos: any): ConnectedPosition => {
      return {
        originX: 'center', originY: pos, overlayX: 'center', overlayY: revert[pos]
      };
    };
    const centerX = (pos: any): ConnectedPosition => {
      return {
        originX: revert[pos], originY: 'center', overlayX: pos, overlayY: 'center'
      };
    };
    const ltr = (axis: any): ConnectedPosition => {
      return {
        originX: 'start', originY: axis, overlayX: 'start', overlayY: revert[axis]
      };
    };
    const rtl = (axis: any): ConnectedPosition => {
      return {
        originX: 'end', originY: axis, overlayX: 'end', overlayY: revert[axis]
      };
    };
    const side = (axis: any, axis2: any): ConnectedPosition => {
      return {
        originX: axis, originY: axis2, overlayX: revert[axis], overlayY: axis2
      };
    };

    const cmp = dir || this.options?.dir || this.dir;
    if (cmp === 'ct') {
      return [centerY('top'), centerY('bottom'), centerX('start'), centerX('end')];
    }
    if (cmp === 'cr') {
      return [centerX('end'), centerX('start'), centerY('top'), centerY('bottom')];
    }
    if (cmp === 'cc') {
      return [centerX('center')];
    }
    if (cmp === 'ltr') {
      return [ltr('bottom'), ltr('top')];
    }
    if (cmp === 'rtl') {
      return [rtl('bottom'), rtl('top')];
    }
    if (cmp === 'tltr') {
      return [ltr('top'), ltr('bottom')];
    }
    if (cmp === 'trtl') {
      return [rtl('top'), rtl('bottom')];
    }
    if (cmp === 'rt') {
      return [side('end', 'top'), side('end', 'bottom'), ltr('top')];
    }
    if (cmp === 'rb') {
      return [side('end', 'bottom'), side('end', 'top'), ltr('bottom'), rtl('bottom')];
    }
    if (cmp === 'lt') {
      return [side('start', 'top'), side('start', 'bottom'), rtl('top')];
    }
    if (cmp === 'lb') {
      return [side('start', 'bottom'), side('start', 'top'), rtl('bottom')];
    }

    return [];
  }

  createPopover(
    origin?: FlexibleConnectedPositionStrategyOrigin,
    options?: MonkeyEcxPopoverOptions
  ) {
    this.templatePortal = new TemplatePortal(this.tpl, this._viewContainerRef);
    this.positionStrategy = this.overlayPositionBuilder
      .flexibleConnectedTo(origin || this.target)
      .withPush(true)
      .withGrowAfterOpen(true)
      .withPositions(this.getDirPositions(options?.dir));

    this.overlayRef = this.overlay.create({
      positionStrategy: this.positionStrategy,
      disposeOnNavigation: true,
      direction: 'ltr',
      hasBackdrop: this.backdrop,
      scrollStrategy: this.overlay.scrollStrategies.reposition({
      }),
      backdropClass: 'cdk-overlay-transparent-backdrop',
      width: (this.options?.minwidth || this.minwidth) ? (this.target.offsetWidth - 2) : undefined,
      maxHeight: '95%',
      height: this.height
    });

    this.subs = this.overlayRef.backdropClick().subscribe((e) => {
      this.onHandleClose(e);
    });

    const attached = this.overlayRef.attach(this.templatePortal);

    if (this.watch) {
      try {
        if ('ResizeObserver' in window) {
          this.resizeObserver = new ResizeObserver(([{ contentRect }]) => {
            this.overlayRef.updatePosition();
          })
            .observe(attached.rootNodes[0]);
        }
      } catch (e) {
        // not to do
      }
    }
  }

  ngOnDestroy() {
    this.popover = false;
    this.subs.unsubscribe();
    this.eventsSubs.unsubscribe();
  }
}
