import { Injectable, NgZone } from '@angular/core';
import { Router } from '@angular/router';
import { HttpErrorResponse } from '@angular/common/http';
import { TranslateService } from '@ngx-translate/core';
import { MonkeyStyleGuideSnackbarService } from 'monkey-style-guide';
import { MonkeyEcxAuthenticationService } from '../auth';
import { MonkeyEcxService } from '../monkeyecx-service.service';
import { MonkeyEcxTokenStorageService } from '../storage';
import { MonkeyEcxCommonsService } from '../commons';
import { MonkeyEcxServiceCredentials } from '../../interfaces';
import { MonkeyEcxUtils } from '../../utils/utils';

@Injectable({
  providedIn: 'root'
})
export class MonkeyEcxHttpErrorHandlingService extends MonkeyEcxCommonsService {
  constructor(
    monkeyecxService: MonkeyEcxService,
    tokenStorage: MonkeyEcxTokenStorageService,
    private monkeyecxAuthenticationService: MonkeyEcxAuthenticationService,
    private router: Router,
    private snackbarService: MonkeyStyleGuideSnackbarService,
    private translateService: TranslateService
  ) {
    super(monkeyecxService, tokenStorage);

    super.resolve(null, null);

    this.geti18n(this.translateService, 'SERVICES.WARNING');
  }

  private isHttpCodeIgnoreMessage(tp: MonkeyEcxServiceCredentials, httpCode: number): boolean {
    if (!MonkeyEcxUtils.persistNullEmptyUndefined(tp) && !tp?.httpResponse) {
      return false;
    }

    const found = (tp?.httpResponse?.httpCodeIgnore || [])?.indexOf(httpCode) > -1;

    return found;
  }

  private isHttpCodeIgnoreRedirect(
    tp: MonkeyEcxServiceCredentials,
    httpCode: number
  ): boolean {
    if (!MonkeyEcxUtils.persistNullEmptyUndefined(tp) && !tp?.httpResponse) {
      return false;
    }

    const found = (tp?.httpResponse?.httpCodeIgnoreRedirect || [])?.indexOf(httpCode) > -1;

    return found;
  }

  private getMessageType(status: number): 'theme' | 'warning' | 'success' | 'error' | 'question' {
    if (status === 409 || status === 400) {
      return 'warning';
    }

    return 'error';
  }

  private showMessage(
    message: string,
    error: HttpErrorResponse,
    mkc: MonkeyEcxServiceCredentials
  ) {
    const { snackbarService } = this;
    if (!this.isHttpCodeIgnoreMessage(mkc, error.status)) {
      snackbarService.show({
        title: this.__i18n?.TITLE,
        message,
        type: this.getMessageType(error.status),
        duration: 4000,
        horizontalPosition: 'center',
        verticalPosition: 'top'
      });
    }
  }

  private handleMessage(error: HttpErrorResponse, mkc: MonkeyEcxServiceCredentials) {
    // eslint-disable-next-line camelcase
    const { notifications, error_description, message } = error.error;
    let customMessage = '';

    if (message) {
      customMessage = message;
    }

    // eslint-disable-next-line camelcase
    if (error_description) {
      // eslint-disable-next-line camelcase
      customMessage = error_description;
    }

    if (notifications) {
      customMessage = notifications;
    }

    if (!customMessage) {
      if (error.error instanceof Blob) {
        const blob = new Blob([error.error], {
          type: 'application/json'
        });
        const reader = new FileReader();
        reader.addEventListener('loadend', (e) => {
          const text = e?.target?.result;
          const jsonResp = JSON.parse(text as string);
          customMessage = jsonResp.notifications;
          this.showMessage(customMessage, error, mkc);
        });
        reader.readAsText(blob);
      } else {
        this.showMessage(this.__i18n['ERROR-HANDLING']?.ERROR, error, mkc);
      }
    } else {
      this.showMessage(customMessage, error, mkc);
    }
  }

  public handleError(error: HttpErrorResponse, mkc: MonkeyEcxServiceCredentials) {
    const { router, tokenStorage } = this;
    const { status } = error;
    const { companyType } = tokenStorage?.getToken() || { companyType: '' };

    const routes: any = {
      403: 'forbidden',
      500: 'service-problems',
      503: 'service-problems'
    };

    let route = '/app';
    if (companyType) {
      route = `${route}/${companyType}/pages`;
    } else {
      route = `${route}/pages`;
    }

    const found = routes[status];
    let timeout = 0;

    if (found && !this.isHttpCodeIgnoreRedirect(mkc, error?.status)) {
      route = `${route}/${found}`.toLowerCase();
      timeout = 800;
      router.navigate([route]);
    }

    setTimeout(() => {
      this.handleMessage(error, mkc);
    }, timeout);
  }

  public handleErrorRefreshToken(error: HttpErrorResponse, mkc: MonkeyEcxServiceCredentials) {
    if (error.status === 401 || error.status === 400) {
      NgZone.call(this.monkeyecxAuthenticationService.redirectLoginWelcomeBack(), {});
    } else if (!this.isHttpCodeIgnoreMessage(mkc, error.status)) {
      this.handleMessage(error, mkc);
    }
  }
}
