import { Injectable } from '@angular/core';
import { Observable, of, BehaviorSubject } from 'rxjs';
import { MonkeyEcxConfigService } from '../config/monkeyecx-config.service';
import { MonkeyEcxUtils } from '../../utils/utils';
import {
  MonkeyEcxConfig,
  MonkeyEcxMeCredentials,
  MonkeyEcxTokenCredentials,
} from '../../interfaces';

@Injectable({
  providedIn: 'root',
})
export class MonkeyEcxTokenStorageService {
  private token: MonkeyEcxTokenCredentials | null;

  private token$: BehaviorSubject<MonkeyEcxTokenCredentials | any>;

  private me$: BehaviorSubject<MonkeyEcxMeCredentials | any>;

  private config: MonkeyEcxConfig | null = null;

  constructor(private monkeyecxConfigService: MonkeyEcxConfigService) {
    this.token = {
      username: '',
      program: '',
      accessToken: '',
      refreshToken: '',
      companyId: '',
      governmentId: '',
      companyTypeUrl: '',
      companyType: '',
      programAdmin: false,
      programType: '',
      accessType: '',
      role: '',
      ownerGovernmentId: '',
      _clearIgnore: {
        ignore: ['username', 'program', 'accessType'],
      },
    };
    this.token$ = new BehaviorSubject(null);
    this.me$ = new BehaviorSubject(null);

    this.monkeyecxConfigService.config().subscribe((_: MonkeyEcxConfig) => {
      this.config = _;
    });
  }

  private setAllTokens(token: MonkeyEcxTokenCredentials) {
    const { token$ } = this;
    Object.entries(token).forEach(([key, value]) => {
      if (MonkeyEcxUtils.persistNullEmptyUndefined(value)) {
        localStorage.setItem(key, value);
      }
    });
    token$.next(token);
  }

  private getAllTokens(): MonkeyEcxTokenCredentials {
    const { token, config } = this;
    const handledToken: any = {
      ...token,
    };
    Object.entries(token || {}).forEach(([key, value]) => {
      handledToken[key] = localStorage.getItem(key);
    });

    if (!handledToken?.program && config?.program) {
      handledToken.program = config?.program?.token;
    }

    return handledToken;
  }

  private setAllMe(me: MonkeyEcxMeCredentials) {
    const { me$ } = this;
    let previousMe = JSON.parse(localStorage?.getItem('me') || '{}');
    Object.entries(me).forEach(([key, value]) => {
      if (MonkeyEcxUtils.persistNullEmptyUndefined(value)) {
        previousMe = {
          ...previousMe,
          [key]: value,
        };
      }
    });

    localStorage.setItem('me', JSON.stringify(previousMe));
    me$.next(previousMe);
  }

  private getAllMe(): MonkeyEcxMeCredentials {
    const me = localStorage.getItem('me');
    return JSON.parse(me || '{}') as MonkeyEcxMeCredentials;
  }

  public tokenHasChanged(): Observable<MonkeyEcxTokenCredentials> {
    return this.token$.asObservable();
  }

  public meHasChanged(): Observable<MonkeyEcxMeCredentials> {
    return this.me$.asObservable() || null;
  }

  public getToken(): MonkeyEcxTokenCredentials {
    return this.getAllTokens();
  }

  public getMe(): MonkeyEcxMeCredentials {
    return this.getAllMe();
  }

  public setToken(token: MonkeyEcxTokenCredentials) {
    this.setAllTokens(token);
  }

  public setMe(me: MonkeyEcxMeCredentials) {
    this.setAllMe(me);
  }

  public clear(force?: boolean): Observable<boolean> {
    const { token } = this;
    try {
      Object.entries(token || {}).forEach(([key, value]) => {
        if (
          (token?._clearIgnore?.ignore || []).filter((val: any) => {
            return val === key;
          }).length <= 0 ||
          force === true
        ) {
          localStorage.removeItem(key);
        }
      });
      localStorage.removeItem('me');
    } catch (error) {
      return of(false);
    }
    return of(true);
  }

  /**
   * @param fields (the name of fields you want to clear. ie: ['companyId', 'role', ...])
   */
  public clearCredentials(
    fields = ['companyId', 'governmentId', 'companyType', 'programAdmin', 'role']
  ) {
    fields?.forEach((key: string) => {
      localStorage.removeItem(key);
    });
  }
}
