import { AbstractControl, ValidationErrors } from '@angular/forms';
import * as moment_ from 'moment';
import { MonkeyEcxUtils } from './utils';

const moment: any = moment_;

const EMAIL_REGEXP = /^[\w-.]+@([\w-]+\.)+[\w-]{1,64}$/;

function isEmptyInputValue(value: any): boolean {
  return value == null || value.length === 0;
}

export function emailValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) {
    return null;
  }
  const test = EMAIL_REGEXP.test(control.value);
  if (test) return null;
  return {
    email: true,
  };
}

export function dateRangeValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) {
    return null;
  }

  const dates = control?.parent?.get('dates')?.value;
  if (
    dates &&
    (!MonkeyEcxUtils.persistNullEmptyUndefined(dates?.startDate) ||
      !MonkeyEcxUtils.persistNullEmptyUndefined(dates?.endDate))
  ) {
    return {
      dateRange: true,
    };
  }

  return null;
}

export function registerValidator(control: AbstractControl, type: string): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) {
    return null;
  }

  if (control && control.value !== `#MONKEY${type}`.toUpperCase()) {
    return {
      invalidUnlockRegister: true,
    };
  }

  return null;
}

export function dateStartEndValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control) {
    return null;
  }

  const { parent } = control;

  const valueStart = parent?.get('dateStart')?.value;
  const valueEnd = parent?.get('dateEnd')?.value;
  let dateStart: Date | null = null;
  let dateEnd: Date | null = null;
  if (valueStart) {
    dateStart = new Date(valueStart);
  }
  if (valueEnd) {
    dateEnd = new Date(valueEnd);
  }

  if (!dateEnd || !dateStart) return null;
  if (dateStart > dateEnd) {
    return {
      dateStartMustBeLessThanAnd: true,
    };
  }

  return null;
}

export function urlValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) return null;

  if (!MonkeyEcxUtils.isValidUrl(control.value)) {
    return {
      url: true,
    };
  }

  return null;
}

export function passwordConfirmValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) return null;

  const { parent } = control;
  const password = parent?.get('password')?.value;
  const passwordConfirm = parent?.get('passwordConfirm')?.value;
  if (!password || !passwordConfirm) return null;
  if (password === passwordConfirm) return null;

  return {
    passwordsNotMatching: true,
  };
}

export function trueValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control) return null;
  if (control && control.value !== false) return null;

  return {
    invalidTrue: true,
  };
}

export function comboValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control) return null;
  if (control && control.value !== '0') return null;

  return {
    invalidCombo: true,
  };
}

export function zipCodeValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) return null;
  if (control && control.value) {
    if (!MonkeyEcxUtils.isValidZipCode(control.value)) {
      return {
        invalidZipCode: true,
      };
    }
  }

  return null;
}

export function documentValidator(
  control: AbstractControl,
  country: string
): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) return null;
  if (country === 'BR') {
    if (!MonkeyEcxUtils.isCPFCNPJValid(control.value)) {
      return {
        invalidCpfCnpj: true,
      };
    }
  } else if (country === 'CL') {
    if (!MonkeyEcxUtils.isValidRUT(control.value)) {
      return {
        invalidCpfCnpj: true,
      };
    }
  }

  return null;
}

export function dateValidator(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) return null;

  const dateFormat = 'MM-DD-YYYY';
  if (
    !moment
      .default(
        moment.default(control.value).format(dateFormat),
        ['DD/MM/YYYY', 'MM-DD-YYYY', 'YYYY-MM-DD'],
        true
      )
      .isValid()
  ) {
    return {
      invalidDate: true,
    };
  }
  return null;
}

export function valueGreaterThanZero(control: AbstractControl): ValidationErrors | null {
  if (!control.parent || !control || isEmptyInputValue(control.value)) return null;

  if (control && `${control?.value}` === '0') {
    return {
      invalidValueGreaterThanZero: true,
    };
  }
  return null;
}

export class Validators {
  static email(control: AbstractControl): ValidationErrors | null {
    return emailValidator(control);
  }

  static dateRange(control: AbstractControl): ValidationErrors | null {
    return dateRangeValidator(control);
  }

  static unlockSponsorRegister(control: AbstractControl): ValidationErrors | null {
    return registerValidator(control, 'sponsor');
  }

  static unlockBuyerRegister(control: AbstractControl): ValidationErrors | null {
    return registerValidator(control, 'buyer');
  }

  static dateStartEnd(control: AbstractControl): ValidationErrors | null {
    return dateStartEndValidator(control);
  }

  static url(control: AbstractControl): ValidationErrors | null {
    return urlValidator(control);
  }

  static passwordConfirm(control: AbstractControl): ValidationErrors | null {
    return passwordConfirmValidator(control);
  }

  static true(control: AbstractControl): ValidationErrors | null {
    return trueValidator(control);
  }

  static combo(control: AbstractControl): ValidationErrors | null {
    return comboValidator(control);
  }

  static zipCode(control: AbstractControl): ValidationErrors | null {
    return zipCodeValidator(control);
  }

  static documentBR(control: AbstractControl): ValidationErrors | null {
    return documentValidator(control, 'BR');
  }

  static documentCL(control: AbstractControl): ValidationErrors | null {
    return documentValidator(control, 'CL');
  }

  static date(control: AbstractControl): ValidationErrors | null {
    return dateValidator(control);
  }

  static greaterThanZero(control: AbstractControl): ValidationErrors | null {
    return valueGreaterThanZero(control);
  }
}
